/*
 * Copyright (c) 2003 Constantin S. Svintsoff <kostik@iclub.nsu.ru>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the authors may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <pthread.h>

#if 1
#include "truepath.c"
#define realpath(a, b)	truepath((a), (b))
#endif

pthread_mutex_t m;
pthread_cond_t c;

void *
thread(void *arg)
{
	const char *name = arg;
	char rname[MAXPATHLEN];
	char dname[MAXPATHLEN];
	int i;

	errno = 0;
	if (mkdir(name, 0777) < 0 && errno != EEXIST) {
		perror(name);
		exit(1);
	}
	strlcpy(rname, name, sizeof(rname));
	strlcat(rname, "/foobar", sizeof(rname));
	errno = 0;
	if (mkdir(rname, 0777) < 0 && errno != EEXIST) {
		perror(name);
		exit(1);
	}

	getcwd(dname, sizeof(dname));
	strlcat(dname, "/", sizeof(dname));
        strlcat(dname, rname, sizeof(dname));
	fprintf(stderr, "%s: %s\n", name, dname);

	/*
	 * wait for start
	 */
	fprintf(stderr, "%s: waiting for main thread to wake us up\n", name);
	pthread_mutex_lock(&m);
	pthread_cond_wait(&c, &m);
	pthread_mutex_unlock(&m);
	fprintf(stderr, "%s: started\n", name);

	for (i = 0; i < 100000; i++) {
		char buf[MAXPATHLEN];

		if (realpath(rname, buf) == 0) {
			perror(buf);
			assert(0);
		}

		if (!!strcmp(buf, dname)) {
			fprintf(stderr, "%s: expected '%s' got '%s'\n",
				name, dname, buf);
			assert(0);
		}
	}

        fprintf(stderr, "%s: finished\n", name);
	rmdir(rname);
	rmdir(name);

	return NULL;
}

void
regress()
{
	char resolved[MAXPATHLEN];
	static const char loop[] = "loop";

	errno = 0;
	if (symlink(loop, loop) < 0 && errno != EEXIST) {
		perror(loop);
		return;
	}
	if (realpath(loop, resolved) == NULL) {
		fprintf(stderr, "realpath: %s: %s\n",
			resolved, strerror(errno));
	}
}

int
main()
{
	pthread_t p1, p2;

	regress();

	pthread_mutex_init(&m, NULL);
	pthread_cond_init(&c, NULL);

	pthread_create(&p1, NULL, thread, "p1");
	pthread_create(&p2, NULL, thread, "p2");

	fprintf(stderr, "main: sleeping for 1 sec\n");
	sleep(1);
	fprintf(stderr, "main: waking up the threads\n");
	pthread_mutex_lock(&m);
	pthread_cond_broadcast(&c);
	pthread_mutex_unlock(&m);

	pthread_join(p1, NULL);
	pthread_join(p2, NULL);

	pthread_cond_destroy(&c);
	pthread_mutex_destroy(&m);

	exit(0);
}
